#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import random

from os import environ
from mezmorize import Cache

if True:
    config = {
        'DEBUG': True,
        'CACHE_TYPE': 'memcached',
        'CACHE_MEMCACHED_SERVERS': [environ.get('MEMCACHE_SERVERS')]}
else:
    config = {'CACHE_TYPE': 'simple'}

cache = Cache(**config)


@cache.memoize(60)
def add(a, b):
    return a + b + random.randrange(0, 1000)


@cache.memoize(60)
def sub(a, b):
    return a - b - random.randrange(0, 1000)


def delete_cache():
    cache.delete_memoized(add)
    cache.delete_memoized(sub)
    return 'caches deleted'


if __name__ == '__main__':
    print('Initial add(2, 5): %s' % add(2, 5))
    print('Memoized add(2, 5): %s' % add(2, 5))
    print('Initial sub(2, 5): %s' % sub(2, 5))
    print('Memoized (sub(2, 5): %s' % sub(2, 5))
    print('Delete all caches')
    delete_cache()
    print('Initial add(2, 5): %s' % add(2, 5))
    print('Initial sub(2, 5): %s' % sub(2, 5))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
    mezmorize.utils
    ~~~~~~~~~~~~~~~

    Provides mezmorize utility functions
"""
from __future__ import absolute_import, division, print_function

import sys

from os import getenv
from subprocess import call
from copy import copy

try:
    import pylibmc
except ImportError:
    pylibmc = None

try:
    import pymemcache
except ImportError:
    pymemcache = None

try:
    import bmemcached
except ImportError:
    bmemcached = None

try:
    import redis
except ImportError:
    redis = None

IS_PY3 = sys.version_info.major == 3

DEF_THRESHOLD = 2048
DEF_DEFAULT_TIMEOUT = 3600
DEF_MC_HOST = DEF_REDIS_HOST = 'localhost'
DEF_MC_PORT = 11211
DEF_REDIS_PORT = 6379

ALL_MEMCACHES = (
    ('pylibmc', pylibmc), ('pymemcache', pymemcache),
    ('bmemcached', bmemcached))

DEF_MC_SERVERS = '{}:{}'.format(DEF_MC_HOST, DEF_MC_PORT)
MC_SERVERS = getenv('MEMCACHIER_SERVERS') or getenv('MEMCACHEDCLOUD_SERVERS')
MC_SERVERS = MC_SERVERS or getenv('MEMCACHE_SERVERS') or DEF_MC_SERVERS
MC_USERNAME = getenv('MEMCACHIER_USERNAME') or getenv('MEMCACHEDCLOUD_USERNAME')
MC_PASSWORD = getenv('MEMCACHIER_PASSWORD') or getenv('MEMCACHEDCLOUD_PASSWORD')

REDIS_HOST = getenv('REDIS_PORT_6379_TCP_ADDR', DEF_REDIS_HOST)
DEF_REDIS_URL = 'redis://{}:{}'.format(REDIS_HOST, DEF_REDIS_PORT)
REDIS_URL = getenv('REDIS_URL') or getenv('REDISTOGO_URL') or DEF_REDIS_URL

CACHE_CONFIGS = {
    'simple': {'CACHE_TYPE': 'simple'},
    'null': {'CACHE_TYPE': 'null'},
    'redis': {'CACHE_TYPE': 'redis', 'CACHE_REDIS_URL': REDIS_URL},
    'filesystem': {
        'CACHE_TYPE': 'filesystem',
        'CACHE_DIR': getenv('CACHE_DIR')
    },
    'memcached': {
        'CACHE_TYPE': 'memcached',
        'CACHE_MEMCACHED_SERVERS': [MC_SERVERS]
    },
    'saslmemcached': {
        'CACHE_TYPE': 'saslmemcached',
        'CACHE_MEMCACHED_SERVERS': [MC_SERVERS],
        'CACHE_MEMCACHED_USERNAME': MC_USERNAME,
        'CACHE_MEMCACHED_PASSWORD': MC_PASSWORD
    },
    'spreadsaslmemcached': {
        'CACHE_TYPE': 'spreadsaslmemcached',
        'CACHE_MEMCACHED_SERVERS': [MC_SERVERS],
        'CACHE_MEMCACHED_USERNAME': MC_USERNAME,
        'CACHE_MEMCACHED_PASSWORD': MC_PASSWORD
    }
}


HEROKU_PROCESSES = {
    'postgres': ['DATABASE_URL'],
    'redis': ['REDIS_URL', 'REDISTOGO_URL'],
    'memcache': ['MEMCACHIER_SERVERS', 'MEMCACHE_SERVERS'],
}


def pgrep(process):
    envs = HEROKU_PROCESSES.get(process, [])
    any_env = any(map(getenv, envs))
    return any_env or call(['pgrep', process]) == 0


HAS_MEMCACHE = (pylibmc or pymemcache or bmemcached) and pgrep('memcache')
AVAIL_MEMCACHES = {k for k, v in ALL_MEMCACHES if HAS_MEMCACHE and v}
HAS_REDIS = redis and pgrep('redis')


def get_cache_type(cache=None, spread=False):
    if HAS_REDIS and HAS_MEMCACHE and not cache:
        cache = 'memcached'
    elif not cache:
        cache = 'redis' if HAS_REDIS else 'memcached'

    if HAS_MEMCACHE and cache == 'memcached':
        if MC_USERNAME and spread:
            cache_type = 'spreadsaslmemcached'
        elif MC_USERNAME:
            cache_type = 'saslmemcached'
        else:
            cache_type = 'memcached'
    elif HAS_REDIS and cache == 'redis':
        cache_type = 'redis'
    elif getenv('CACHE_DIR'):
        cache_type = 'filesystem'
    else:
        cache_type = 'simple'

    return cache_type


def get_cache_config(cache_type, db=None, **kwargs):
    config = copy(CACHE_CONFIGS[cache_type])
    redis_url = config.get('CACHE_REDIS_URL')

    if db and redis_url:
        config['CACHE_REDIS_URL'] = '{}/{}'.format(redis_url, db)

    options = {k: v for k, v in kwargs.items() if v is not None}
    options.setdefault('CACHE_THRESHOLD', DEF_THRESHOLD)
    options.setdefault('CACHE_DEFAULT_TIMEOUT', DEF_DEFAULT_TIMEOUT)
    config.update(options)
    return config


def get_pylibmc_client(servers, timeout=None, binary=True, **kwargs):
    from pylibmc import Client

    try:
        from pylibmc import TooBig
    except ImportError:
        from pylibmc import Error, ServerError
        TooBig = (Error, ServerError)

    if timeout:
        kwargs['behaviors'] = {'connect_timeout': timeout}

    client = Client(servers, binary=binary, **kwargs)
    client.TooBig = TooBig
    return client


def get_pymemcache_client(servers, timeout=None, **kwargs):
    from pymemcache.client.hash import HashClient

    from pymemcache.serde import (
        python_memcache_serializer, python_memcache_deserializer)

    kwargs.setdefault('serializer', python_memcache_serializer)
    kwargs.setdefault('deserializer', python_memcache_deserializer)

    if timeout:
        kwargs['timeout'] = timeout

    split = [s.split(':') for s in servers]
    _servers = [(host, int(port)) for host, port in split]
    client = HashClient(_servers, **kwargs)

    try:
        client.TooBig = ConnectionResetError
    except NameError:
        import socket
        client.TooBig = socket.error

    return client


def get_bmemcached_client(servers, timeout=None, **kwargs):
    from bmemcached import Client

    if timeout:
        kwargs['socket_timeout'] = timeout

    client = Client(servers, **kwargs)
    client.TooBig = None
    return client



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
    tests.test_cache
    ~~~~~~~~~~~~~~~~

    Provides unit tests.
"""
from __future__ import absolute_import, division, print_function

import time
import random

import nose.tools as nt

from mezmorize import Cache, function_namespace
from mezmorize.utils import HAS_MEMCACHE, HAS_REDIS, get_cache_config

from mezmorize.backends import (
    SimpleCache, FileSystemCache, RedisCache, MemcachedCache,
    SASLMemcachedCache, SpreadSASLMemcachedCache, AVAIL_MEMCACHES)

BIGINT = 2 ** 21
BIGGERINT = 2 ** 28


def setup_func(*args, **kwargs):
    namespace = kwargs.pop('namespace', None)
    client_name = kwargs.pop('client_name', None)

    if client_name:
        CACHE_OPTIONS = kwargs.get('CACHE_OPTIONS', {})
        CACHE_OPTIONS['preferred_memcache'] = client_name
        kwargs['CACHE_OPTIONS'] = CACHE_OPTIONS

    config = get_cache_config(*args, **kwargs)
    cache = Cache(namespace=namespace, **config)
    return cache


def check_cache_type(cache, cache_type):
    nt.assert_equal(cache.config['CACHE_TYPE'], cache_type)


def check_cache_instance(cache, cache_instance):
    nt.assert_is_instance(cache.cache, cache_instance)


def check_client_name(cache, expected):
    nt.assert_equal(cache.cache.client_name, expected)


def check_too_big(cache, times, error=None):
    if cache.cache.TooBig:
        with nt.assert_raises(error or cache.cache.TooBig):
            cache.set('big', 'a' * times)

        nt.assert_is_none(cache.get('big'))
    else:
        cache.set('big', 'a' * times)
        nt.assert_equal(cache.get('big'), 'a' * times)


def check_set_delete(cache, key, value, multiplier=None):
    if multiplier:
        value *= multiplier

    cache.set(key, value)
    nt.assert_equal(cache.get(key), value)

    cache.delete(key)
    nt.assert_is_none(cache.get(key))


class TestCache(object):
    def setup(self):
        self.cache = setup_func('simple')

    def teardown(self):
        self.cache.clear()

    def test_dict_config(self):
        check_cache_type(self.cache, 'simple')
        check_cache_instance(self.cache, SimpleCache)

    def test_000_set(self):
        self.cache.set('hi', 'hello')
        nt.assert_equal(self.cache.get('hi'), 'hello')

    def test_add(self):
        self.cache.add('hi', 'hello')
        nt.assert_equal(self.cache.get('hi'), 'hello')

        self.cache.add('hi', 'foobar')
        nt.assert_equal(self.cache.get('hi'), 'hello')

    def test_add_unicode(self):
        self.cache.add('ȟį', 'ƕɛĺłö')
        nt.assert_equal(self.cache.get('ȟį'), 'ƕɛĺłö')

        self.cache.add('ȟį', 'fööƀåř')
        nt.assert_equal(self.cache.get('ȟį'), 'ƕɛĺłö')

    def test_delete(self):
        check_set_delete(self.cache, 'hi', 'hello')

    def test_delete_unicode(self):
        check_set_delete(self.cache, 'ȟį', 'ƕɛĺłö')

    def test_memoize(self):
        @self.cache.memoize(5)
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        result = func(5, 2)
        time.sleep(1)
        nt.assert_equal(func(5, 2), result)

        result2 = func(5, 3)
        nt.assert_not_equal(result2, result)

        time.sleep(6)
        nt.assert_not_equal(func(5, 2), result)

        time.sleep(1)
        nt.assert_not_equal(func(5, 3), result2)

    def test_timeout(self):
        config = get_cache_config('simple', CACHE_DEFAULT_TIMEOUT=1)
        self.cache = Cache(**config)

        @self.cache.memoize(50)
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        result = func(5, 2)
        time.sleep(2)
        nt.assert_equal(func(5, 2), result)

    def test_delete_timeout(self):
        @self.cache.memoize(5)
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        result = func(5, 2)
        result2 = func(5, 3)
        time.sleep(1)

        nt.assert_equal(func(5, 2), result)
        nt.assert_equal(func(5, 2), result)
        nt.assert_not_equal(func(5, 3), result)
        nt.assert_equal(func(5, 3), result2)

        self.cache.delete_memoized(func)
        nt.assert_not_equal(func(5, 2), result)
        nt.assert_not_equal(func(5, 3), result2)

    def test_delete_verhash(self):
        @self.cache.memoize(5)
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        result = func(5, 2)
        result2 = func(5, 3)
        time.sleep(1)

        nt.assert_equal(func(5, 2), result)
        nt.assert_equal(func(5, 2), result)
        nt.assert_not_equal(func(5, 3), result)
        nt.assert_equal(func(5, 3), result2)

        fname = function_namespace(func)[0]
        version_key = self.cache._memvname(fname)
        nt.assert_is_not_none(self.cache.get(version_key))

        self.cache.delete_memoized_verhash(func)
        nt.assert_is_none(self.cache.get(version_key))
        nt.assert_not_equal(func(5, 2), result)
        nt.assert_not_equal(func(5, 3), result2)
        nt.assert_is_not_none(self.cache.get(version_key))

    def test_delete_rand(self):
        @self.cache.memoize()
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        result_a = func(5, 1)
        result_b = func(5, 2)
        nt.assert_equal(func(5, 1), result_a)
        nt.assert_equal(func(5, 2), result_b)

        self.cache.delete_memoized(func, 5, 2)
        nt.assert_equal(func(5, 1), result_a)
        nt.assert_not_equal(func(5, 2), result_b)

    def test_args(self):
        @self.cache.memoize()
        def func(a, b):
            return sum(a) + sum(b) + random.randrange(0, 100000)

        result_a = func([5, 3, 2], [1])
        result_b = func([3, 3], [3, 1])
        nt.assert_equal(func([5, 3, 2], [1]), result_a)
        nt.assert_equal(func([3, 3], [3, 1]), result_b)

        self.cache.delete_memoized(func, [5, 3, 2], [1])
        nt.assert_not_equal(func([5, 3, 2], [1]), result_a)
        nt.assert_equal(func([3, 3], [3, 1]), result_b)

    def test_kwargs(self):
        @self.cache.memoize()
        def func(a, b=None):
            return a + sum(b.values()) + random.randrange(0, 100000)

        result_a = func(1, {'one': 1, 'two': 2})
        result_b = func(5, {'three': 3, 'four': 4})
        nt.assert_equal(func(1, {'one': 1, 'two': 2}), result_a)
        nt.assert_equal(func(5, {'three': 3, 'four': 4}), result_b)

        self.cache.delete_memoized(func, 1, {'one': 1, 'two': 2})
        nt.assert_not_equal(func(1, {'one': 1, 'two': 2}), result_a)
        nt.assert_equal(func(5, {'three': 3, 'four': 4}), result_b)

    def test_kwargonly(self):
        @self.cache.memoize()
        def func(a=None):
            if a is None:
                a = 0
            return a + random.random()

        result_a = func()
        result_b = func(5)

        nt.assert_equal(func(), result_a)
        nt.assert_less(func(), 1)
        nt.assert_equal(func(5), result_b)
        nt.assert_greater_equal(func(5), 5)
        nt.assert_less(func(5), 6)

    def test_arg_kwarg(self):
        @self.cache.memoize()
        def func(a, b, c=1):
            return a + b + c + random.randrange(0, 100000)

        nt.assert_equal(func(1, 2), func(1, 2, c=1))
        nt.assert_equal(func(1, 2), func(1, 2, 1))
        nt.assert_equal(func(1, 2), func(1, 2))
        nt.assert_not_equal(func(1, 2, 3), func(1, 2))

        with nt.assert_raises(TypeError):
            func(1)

    def test_classarg(self):
        @self.cache.memoize()
        def func(a):
            return a.value + random.random()

        class Adder(object):
            def __init__(self, value):
                self.value = value

        adder = Adder(15)
        adder2 = Adder(20)

        y = func(adder)
        z = func(adder2)
        nt.assert_not_equal(y, z)
        nt.assert_equal(func(adder), y)
        nt.assert_not_equal(func(adder), z)

        adder.value = 14
        nt.assert_equal(func(adder), y)
        nt.assert_not_equal(func(adder), z)
        nt.assert_not_equal(func(adder), func(adder2))
        nt.assert_equal(func(adder2), z)

    def test_classfunc(self):
        class Adder(object):
            def __init__(self, initial):
                self.initial = initial

            @self.cache.memoize()
            def add(self, b):
                return self.initial + b

        adder1 = Adder(1)
        adder2 = Adder(2)

        x = adder1.add(3)
        nt.assert_equal(adder1.add(3), x)
        nt.assert_not_equal(adder1.add(4), x)
        nt.assert_not_equal(adder1.add(3), adder2.add(3))

    def test_delete_classfunc(self):
        class Adder(object):
            def __init__(self, initial):
                self.initial = initial

            @self.cache.memoize()
            def add(self, b):
                return self.initial + b + random.random()

        adder1 = Adder(1)
        adder2 = Adder(2)

        a1 = adder1.add(3)
        a2 = adder2.add(3)
        nt.assert_not_equal(a1, a2)
        nt.assert_equal(adder1.add(3), a1)
        nt.assert_equal(adder2.add(3), a2)

        self.cache.delete_memoized(adder1.add)
        a3 = adder1.add(3)
        a4 = adder2.add(3)

        nt.assert_not_equal(a1, a3)
        nt.assert_not_equal(a1, a3)
        nt.assert_equal(a2, a4)

        self.cache.delete_memoized(Adder.add)
        a5 = adder1.add(3)
        a6 = adder2.add(3)

        nt.assert_not_equal(a5, a6)
        nt.assert_not_equal(a3, a5)
        nt.assert_not_equal(a4, a6)

    def test_delete_classmethod(self):
        class Mock(object):
            @classmethod
            @self.cache.memoize(5)
            def func(cls, a, b):
                return a + b + random.randrange(0, 100000)

        result = Mock.func(5, 2)
        result2 = Mock.func(5, 3)
        time.sleep(1)

        nt.assert_equal(Mock.func(5, 2), result)
        nt.assert_equal(Mock.func(5, 2), result)
        nt.assert_not_equal(Mock.func(5, 3), result)
        nt.assert_equal(Mock.func(5, 3), result2)

        self.cache.delete_memoized(Mock.func)
        nt.assert_not_equal(Mock.func(5, 2), result)
        nt.assert_not_equal(Mock.func(5, 3), result2)

    def test_multiple_arg_kwarg_calls(self):
        @self.cache.memoize()
        def func(a, b, c=[1, 1], d=[1, 1]):
            rand = random.randrange(0, 100000)
            return sum(a) + sum(b) + sum(c) + sum(d) + rand

        expected = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        nt.assert_equal(func([5, 3, 2], [1], d=[3, 3], c=[3, 3]), expected)

        result = func(b=[1], a=[5, 3, 2], c=[3, 3], d=[3, 3])
        nt.assert_equal(result, expected)
        nt.assert_equal(func([5, 3, 2], [1], [3, 3], [3, 3]), expected)

    def test_delete_multiple_arg_kwarg(self):
        @self.cache.memoize()
        def func(a, b, c=[1, 1], d=[1, 1]):
            rand = random.randrange(0, 100000)
            return sum(a) + sum(b) + sum(c) + sum(d) + rand

        result_a = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        self.cache.delete_memoized(func, [5, 3, 2], [1], [3, 3], [3, 3])
        result_b = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        nt.assert_not_equal(result_a, result_b)

        self.cache.delete_memoized(
            func, [5, 3, 2], b=[1], c=[3, 3], d=[3, 3])
        result_b = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        nt.assert_not_equal(result_a, result_b)

        self.cache.delete_memoized(func, [5, 3, 2], [1], c=[3, 3], d=[3, 3])
        result_a = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        nt.assert_not_equal(result_a, result_b)

        self.cache.delete_memoized(
            func, [5, 3, 2], b=[1], c=[3, 3], d=[3, 3])
        result_a = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        nt.assert_not_equal(result_a, result_b)

        self.cache.delete_memoized(func, [5, 3, 2], [1], c=[3, 3], d=[3, 3])
        result_b = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        nt.assert_not_equal(result_a, result_b)

        self.cache.delete_memoized(func, [5, 3, 2], [1], [3, 3], [3, 3])
        result_a = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        nt.assert_not_equal(result_a, result_b)

    def test_kwargs_to_args(self):
        def func(a, b, c=None, d=None):
            return sum(a) + sum(b) + random.randrange(0, 100000)

        expected = (1, 2, 'foo', 'bar')

        args = self.cache._gen_args(func, 1, 2, 'foo', 'bar')
        nt.assert_equal(tuple(args), expected)

        args = self.cache._gen_args(func, 2, 'foo', 'bar', a=1)
        nt.assert_equal(tuple(args), expected)

        args = self.cache._gen_args(func, a=1, b=2, c='foo', d='bar')
        nt.assert_equal(tuple(args), expected)

        args = self.cache._gen_args(func, d='bar', b=2, a=1, c='foo')
        nt.assert_equal(tuple(args), expected)

        args = self.cache._gen_args(func, 1, 2, d='bar', c='foo')
        nt.assert_equal(tuple(args), expected)


class TestNSCache(object):
    def setup(self):
        self.namespace = 'https://github.com/reubano/mezmorize'
        self.cache = setup_func('simple', namespace=self.namespace)

    def teardown(self):
        self.cache.clear()

    def test_memoize(self):
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        config = get_cache_config('simple')
        cache = Cache(namespace=self.namespace, **config)
        cache_key1 = self.cache._memoize_make_cache_key()(func)
        cache_key2 = cache._memoize_make_cache_key()(func)
        nt.assert_equal(cache_key1, cache_key2)


class TestFileSystemCache(TestCache):
    def setup(self):
        self.cache = setup_func('filesystem', CACHE_DIR='/tmp')

    def teardown(self):
        self.cache.clear()

    def test_dict_config(self):
        check_cache_type(self.cache, 'filesystem')
        check_cache_instance(self.cache, FileSystemCache)


if HAS_MEMCACHE:
    class TestMemcachedCache(TestCache):
        def setup(self, client_name=None):
            self.cache = setup_func('memcached', client_name=client_name)

        def teardown(self):
            self.cache.clear()

        def test_dict_config(self):
            for client_name in AVAIL_MEMCACHES:
                self.setup(client_name=client_name)
                yield check_cache_type, self.cache, 'memcached'
                yield check_cache_instance, self.cache, MemcachedCache
                yield check_client_name, self.cache, client_name
                self.teardown()

        def test_mc_large_value(self):
            for client_name in AVAIL_MEMCACHES:
                self.setup(client_name=client_name)
                yield check_too_big, self.cache, BIGINT
                self.teardown()

else:
    print('TestMemcachedCache requires Memcache')

if HAS_MEMCACHE:
    class TestSASLMemcachedCache(TestCache):
        def setup(self, client_name=None):
            self.cache = setup_func('saslmemcached', client_name=client_name)

        def teardown(self):
            self.cache.clear()

        def test_dict_config(self):
            for client_name in AVAIL_MEMCACHES:
                self.setup(client_name=client_name)
                yield check_cache_type, self.cache, 'saslmemcached'
                yield check_cache_instance, self.cache, SASLMemcachedCache
                self.teardown()

        def test_mc_large_value(self):
            for client_name in AVAIL_MEMCACHES:
                self.setup(client_name=client_name)
                yield check_too_big, self.cache, BIGINT
                self.teardown()

else:
    print('TestSASLMemcachedCache requires Memcache')

if HAS_MEMCACHE:
    class TestSpreadSASLMemcachedCache(TestCache):
        def setup(self, client_name=None):
            cache_type = 'spreadsaslmemcached'
            self.cache = setup_func(cache_type, client_name=client_name)

        def teardown(self):
            self.cache.clear()

        def test_dict_config(self):
            for client_name in AVAIL_MEMCACHES:
                self.setup(client_name=client_name)
                cache_instance = SpreadSASLMemcachedCache
                yield check_cache_type, self.cache, 'spreadsaslmemcached'
                yield check_cache_instance, self.cache, cache_instance
                self.teardown()

        def test_mc_large_value(self):
            for client_name in AVAIL_MEMCACHES:
                self.setup(client_name=client_name)
                yield check_set_delete, self.cache, 'big', 'a', BIGINT
                yield check_too_big, self.cache, BIGGERINT, ValueError
                self.teardown()
else:
    print('TestSpreadSASLMemcachedCache requires Memcache')

if HAS_REDIS:
    class TestRedisCache(TestCache):
        def setup(self, db=0):
            self.cache = setup_func('redis', db=db)
            self.client = self.cache.cache._client

        def teardown(self):
            self.cache.clear()

        def test_dict_config(self):
            check_cache_type(self.cache, 'redis')
            check_cache_instance(self.cache, RedisCache)

        def test_redis_url_default_db(self):
            rconn = self.client.connection_pool.get_connection('foo')
            nt.assert_equal(rconn.db, 0)

        def test_redis_url_custom_db(self):
            self.setup(db=2)
            rconn = self.client.connection_pool.get_connection('foo')
            nt.assert_equal(rconn.db, 2)
else:
    print('TestRedisCache requires Redis')

if __name__ == '__main__':
    nt.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
    mezmorize.backends
    ~~~~~~~~~~~~~~~~~~

    Provides mezmorize storage mechanisms
"""
# pylint: disable=range-builtin-not-iterating

from __future__ import absolute_import, division, print_function

import pickle

from itertools import chain

from werkzeug.contrib.cache import (
    NullCache, SimpleCache, MemcachedCache as _MemcachedCache, FileSystemCache,
    RedisCache)

from .utils import (
    DEF_MC_SERVERS, IS_PY3, HAS_MEMCACHE, AVAIL_MEMCACHES, get_pylibmc_client,
    get_pymemcache_client, get_bmemcached_client, DEF_REDIS_HOST,
    DEF_REDIS_PORT, DEF_DEFAULT_TIMEOUT)

try:
    from redis import from_url
except ImportError:
    from_url = None

CONFIG_LOOKUP = {
    'servers': 'CACHE_MEMCACHED_SERVERS',
    'threshold': 'CACHE_THRESHOLD',
    'timeout': 'CACHE_TIMEOUT',
    'username': 'CACHE_MEMCACHED_USERNAME',
    'password': 'CACHE_MEMCACHED_PASSWORD',
    'key_prefix': 'CACHE_KEY_PREFIX'}


def gen_defaults(*keys, **config):
    for key in keys:
        config_key = CONFIG_LOOKUP[key]

        if config_key in config:
            yield (key, config[config_key])


def get_mc_client(module_name, binary=True, **kwargs):
    servers = kwargs.pop('servers', (DEF_MC_SERVERS,))
    timeout = kwargs.pop('timeout', None)

    if module_name == 'pylibmc':
        client = get_pylibmc_client(
            servers, timeout=timeout, binary=binary, **kwargs)
    elif module_name == 'pymemcache':
        client = get_pymemcache_client(servers, timeout=timeout, **kwargs)
    elif module_name == 'bmemcached':
        client = get_bmemcached_client(servers, timeout=timeout, **kwargs)

    return client


class MemcachedCache(_MemcachedCache):
    def __init__(self, *args, **kwargs):
        default_timeout = kwargs.pop('default_timeout', DEF_DEFAULT_TIMEOUT)
        key_prefix = kwargs.pop('key_prefix', None)

        if not HAS_MEMCACHE:
            raise RuntimeError('No memcache module found.')

        compat_memcaches = kwargs.pop('compat_memcaches', AVAIL_MEMCACHES)
        avail_memcaches = AVAIL_MEMCACHES.intersection(compat_memcaches)

        if not avail_memcaches:
            raise RuntimeError('No compatible memcache module found.')

        preferred_mc = kwargs.pop('preferred_memcache', 'pylibmc')

        if len(avail_memcaches) == 1 or preferred_mc not in avail_memcaches:
            preferred_mc = avail_memcaches.pop()

        client = get_mc_client(preferred_mc, **kwargs)
        skwargs = {'default_timeout': default_timeout, 'key_prefix': key_prefix}
        super(MemcachedCache, self).__init__(servers=client, **skwargs)
        self.TooBig = client.TooBig
        self.client_name = preferred_mc


class SASLMemcachedCache(MemcachedCache):
    def __init__(self, *args, **kwargs):
        kwargs['compat_memcaches'] = ('pylibmc', 'bmemcached')
        super(SASLMemcachedCache, self).__init__(*args, **kwargs)


def null(config, *args, **kwargs):
    return NullCache()


def simple(config, *args, **kwargs):
    defaults = dict(gen_defaults('threshold', 'timeout', **config))
    defaults.update(kwargs)
    return SimpleCache(*args, **kwargs)


def memcached(config, *args, **kwargs):
    keys = ('timeout', 'servers', 'key_prefix')
    defaults = dict(gen_defaults(*keys, **config))
    defaults.update(kwargs)
    return MemcachedCache(*args, **kwargs)


def saslmemcached(config, **kwargs):
    keys = ('timeout', 'servers', 'username', 'password', 'key_prefix')
    defaults = dict(gen_defaults(*keys, **config))
    defaults.update(kwargs)
    return SASLMemcachedCache(**kwargs)


def filesystem(config, *args, **kwargs):
    args = chain([config['CACHE_DIR']], args)
    defaults = dict(gen_defaults('threshold', 'timeout', **config))
    defaults.update(kwargs)
    return FileSystemCache(*args, **kwargs)


def redis(config, *args, **kwargs):
    kwargs.setdefault('host', config.get('CACHE_REDIS_HOST', DEF_REDIS_HOST))
    kwargs.setdefault('port', config.get('CACHE_REDIS_PORT', DEF_REDIS_PORT))
    kwargs.setdefault('password', config.get('CACHE_REDIS_PASSWORD'))
    kwargs.setdefault('key_prefix', config.get('CACHE_KEY_PREFIX'))
    kwargs.setdefault('db', config.get('CACHE_REDIS_DB'))
    redis_url = config.get('CACHE_REDIS_URL')

    if redis_url:
        kwargs['host'] = from_url(redis_url, db=kwargs.pop('db', None))

    return RedisCache(*args, **kwargs)


class SpreadSASLMemcachedCache(SASLMemcachedCache):
    """
    Simple Subclass of SASLMemcached client that spread value across multiple
    key is they are bigger than a given threshold.

    Spreading require using pickle to store the value, which can significantly
    impact the performances.
    """
    DEF_CHUNKSIZE = 2 ** 20 - 2 ** 7  # 1048448
    DEF_MAXCHUNKS = 32

    def __init__(self, *args, **kwargs):
        """
        Kwargs:
            chunksize (int): max length of a pickled object that can fit in
                memcached (memcache has an upper limit of 1MB for values,
                default: 1048448)
        """
        self.CHUNKSIZE = kwargs.get('chunksize', self.DEF_CHUNKSIZE)
        self.MARKER = 'SpreadSASLMemcachedCache.SpreadedValue'
        self.MAXCHUNKS = kwargs.get('maxchunks', self.DEF_MAXCHUNKS)
        super(SpreadSASLMemcachedCache, self).__init__(*args, **kwargs)
        self.super = super(SpreadSASLMemcachedCache, self)

    def _genkeys(self, key):
        return ('{}.{}'.format(key, i) for i in range(self.MAXCHUNKS))

    def _gen_kv(self, key, pickled):
        chunks = range(0, len(pickled), self.CHUNKSIZE)

        if len(chunks) > self.MAXCHUNKS:
            msg = 'Value exceed maximum number of keys ({})'
            raise ValueError(msg.format(self.MAXCHUNKS))

        for i in chunks:
            _key = '{}.{}'.format(key, i // self.CHUNKSIZE)
            _value = pickled[i:i + self.CHUNKSIZE]
            yield _key, _value

    def delete(self, key):
        value = self.super.get(key)
        self.super.delete(key)

        if value == self.MARKER:
            for skey in self._genkeys(key):
                self.super.delete(skey)

    def set(self, key, value, timeout=None):
        """set a value in cache, potentially spreading it across multiple key.
        """
        try:
            value = self.super.set(key, value, timeout=timeout)
        except self.TooBig:
            self.super.set(key, self.MARKER, timeout=timeout)
            pickled = pickle.dumps(value, 2)
            values = dict(self._gen_kv(key, pickled))
            value = self.super.set_many(values, timeout)

        return value

    def get(self, key):
        """get a value in cache, potentially from multiple keys.
        """
        value = self.super.get(key)

        if value == self.MARKER:
            keys = self._genkeys(key)
            result = self.super.get_many(*keys)
            filtered = (v for v in result if v is not None)

            if IS_PY3:
                serialized = b''.join(filtered)
            else:
                serialized = ''.join(filtered)

            value = pickle.loads(serialized) if serialized else None

        return value


def spreadsaslmemcached(config, *args, **kwargs):
    keys = ('timeout', 'servers', 'username', 'password', 'key_prefix')
    defaults = dict(gen_defaults(*keys, **config))
    defaults.update(kwargs)
    return SpreadSASLMemcachedCache(*args, **kwargs)
